#!/usr/bin/env python
#
# Copyright (C) Nathaniel Smith <njs@pobox.com>
# Licensed under the MIT license:
#   http://www.opensource.org/licenses/mit-license.html
# I.e., do what you like, but keep copyright and there's NO WARRANTY.
#
# CIA bot client script for Monotone repositories, written in python.  This
# generates commit messages using CIA's XML commit format, and can deliver
# them using either XML-RPC or email.  Based on the script 'ciabot_svn.py' by
# Micah Dowty <micah@navi.cx>.

# This script is normally run from a cron job.  It periodically does a 'pull'
# from a given server, finds new revisions, filters them for "interesting"
# ones, and reports them to CIA.

# It needs a working directory, where it will store the database and some
# state of its own.

# To use:
#   -- make a copy of it somewhere
#   -- edit the configuration values below
#   -- set up a cron job to run every ten minutes (or whatever), running the
#      command "ciabot_monotone.py <path to scratch dir>".  The scratch dir is
#      used to store state between runs.  It will be automatically created,
#      but do not delete it.

class config:
    def project_for_branch(self, branchname):
        # Customize this to return your project name(s).  If changes to the
        # given branch are uninteresting -- i.e., changes to them should be
        # ignored entirely -- then return the python constant None (which is
        # distinct from the string "None", a valid but poor project name!).
        #if branchname.startswith("net.venge.monotone-viz"):
        #    return "monotone-viz"
        #elif branchname.startswith("net.venge.monotone.contrib.monotree"):
        #    return "monotree"
        #else:
        #    return "monotone"
        return "FIXME"

    # Add entries of the form ("server address", "pattern") to get
    # this script to watch the given collections at the given monotone
    # servers.
    watch_list = [
        #("monotone.ca", "net.venge.monotone"),
        ]

    # If this is non-None, then the web interface will make any file 'foo' a
    # link to 'repository_uri/foo'.
    repository_uri = None

    # The server to deliver XML-RPC messages to, if using XML-RPC delivery.
    xmlrpc_server = "http://cia.navi.cx"

    # The email address to deliver messages to, if using email delivery.
    smtp_address = "cia@cia.navi.cx"

    # The SMTP server to connect to, if using email delivery.
    smtp_server = "localhost"

    # The 'from' address to put on email, if using email delivery.
    from_address = "cia-user@FIXME"

    # Set to one of "xmlrpc", "email", "debug".
    delivery = "debug"

    # Path to monotone executable.
    monotone_exec = "monotone"

################################################################################

import sys, os, os.path

class Monotone:
    def __init__(self, bin, db):
        self.bin = bin
        self.db = db

    def _run_monotone(self, args):
        args_str = " ".join(args)
        # Yay lack of quoting
        fd = os.popen("%s --db=%s --quiet %s" % (self.bin, self.db, args_str))
        output = fd.read()
        if fd.close():
            sys.exit("monotone exited with error")
        return output

    def _split_revs(self, output):
        if output:
            return output.strip().split("\n")
        else:
            return []

    def get_interface_version(self):
        iv_str = self._run_monotone(["automate", "interface_version"])
        return tuple(map(int, iv_str.strip().split(".")))

    def db_init(self):
        self._run_monotone(["db", "init"])

    def db_migrate(self):
        self._run_monotone(["db", "migrate"])

    def ensure_db_exists(self):
        if not os.path.exists(self.db):
            self.db_init()

    def pull(self, server, collection):
        self._run_monotone(["pull", server, collection])

    def leaves(self):
        return self._split_revs(self._run_monotone(["automate", "leaves"]))

    def ancestry_difference(self, new_rev, old_revs):
        args = ["automate", "ancestry_difference", new_rev] + old_revs
        return self._split_revs(self._run_monotone(args))

    def log(self, rev, xlast=None):
        if xlast is not None:
            last_arg = ["--last=%i" % (xlast,)]
        else:
            last_arg = []
        return self._run_monotone(["log", "-r", rev] + last_arg)

    def toposort(self, revs):
        args = ["automate", "toposort"] + revs
        return self._split_revs(self._run_monotone(args))

    def get_revision(self, rid):
        return self._run_monotone(["automate", "get_revision", rid])

class LeafFile:
    def __init__(self, path):
        self.path = path

    def get_leaves(self):
        if os.path.exists(self.path):
            f = open(self.path, "r")
            lines = []
            for line in f:
                lines.append(line.strip())
            return lines
        else:
            return []

    def set_leaves(self, leaves):
        f = open(self.path + ".new", "w")
        for leaf in leaves:
            f.write(leaf + "\n")
        f.close()
        os.rename(self.path + ".new", self.path)

def escape_for_xml(text, is_attrib=0):
    text = text.replace("&", "&amp;")
    text = text.replace("<", "&lt;")
    text = text.replace(">", "&gt;")
    if is_attrib:
        text = text.replace("'", "&apos;")
        text = text.replace("\"", "&quot;")
    return text

def send_message(message, c):
    if c.delivery == "debug":
        print message
    elif c.delivery == "xmlrpc":
        import xmlrpclib
        xmlrpclib.ServerProxy(c.xmlrpc_server).hub.deliver(message)
    elif c.delivery == "email":
        import smtplib
        smtp = smtplib.SMTP(c.smtp_server)
        smtp.sendmail(c.from_address, c.smtp_address,
                      "From: %s\r\nTo: %s\r\n"
                      "Subject: DeliverXML\r\n\r\n%s"
                      % (c.from_address, c.smtp_address, message))
    else:
        sys.exit("delivery option must be one of 'debug', 'xmlrpc', 'email'")

def send_change_for(rid, m, c):
    message_tmpl = """<message>
    <generator>
        <name>Monotone CIA Bot client python script</name>
        <version>0.1</version>
    </generator>
    <source>
        <project>%(project)s</project>
        <branch>%(branch)s</branch>
    </source>
    <body>
        <commit>
            <revision>%(rid)s</revision>
            <author>%(author)s</author>
            <files>%(files)s</files>
            <log>%(log)s</log>
        </commit>
    </body>
</message>"""

    substs = {}

    log = m.log(rid, 1)
    rev = m.get_revision(rid)
    # Stupid way to pull out everything inside quotes (which currently
    # uniquely identifies filenames inside a changeset).
    pieces = rev.split('"')
    files = []
    for i in range(len(pieces)):
        if (i % 2) == 1:
            if pieces[i] not in files:
                files.append(pieces[i])
    substs["files"] = "\n".join(["<file>%s</file>" % escape_for_xml(f) for f in files])
    branch = None
    author = None
    changelog_pieces = []
    started_changelog = 0
    pieces = log.split("\n")
    for p in pieces:
        if p.startswith("Author:"):
            author = p.split(None, 1)[1].strip()
        if p.startswith("Branch:"):
            branch = p.split()[1]
        if p.startswith("ChangeLog:"):
            started_changelog = 1
        elif started_changelog:
            changelog_pieces.append(p)
    changelog = "\n".join(changelog_pieces).strip()
    if branch is None:
        return
    project = c.project_for_branch(branch)
    if project is None:
        return
    substs["author"] = escape_for_xml(author or "(unknown author)")
    substs["project"] = escape_for_xml(project)
    substs["branch"] = escape_for_xml(branch)
    substs["rid"] = escape_for_xml(rid)
    substs["log"] = escape_for_xml(changelog)

    message = message_tmpl % substs
    send_message(message, c)

def send_changes_between(old_leaves, new_leaves, m, c):
    if not old_leaves:
        # Special case for initial setup -- don't push thousands of old
        # revisions down CIA's throat!
        return
    new_revs = {}
    for leaf in new_leaves:
        if leaf in old_leaves:
            continue
        for new_rev in m.ancestry_difference(leaf, old_leaves):
            new_revs[new_rev] = None
    new_revs_sorted = m.toposort(new_revs.keys())
    for new_rev in new_revs_sorted:
        send_change_for(new_rev, m, c)

def main(progname, args):
    if len(args) != 1:
        sys.exit("Usage: %s STATE-DIR" % (progname,))
    (state_dir,) = args
    if not os.path.isdir(state_dir):
        os.makedirs(state_dir)
    lockfile = os.path.join(state_dir, "lock")
    # Small race condition, oh well.
    if os.path.exists(lockfile):
        sys.exit("script already running, exiting")
    try:
        open(lockfile, "w").close()
        c = config()
        m = Monotone(c.monotone_exec, os.path.join(state_dir, "database.db"))
        m.ensure_db_exists()
        m.db_migrate()
        for server, collection in c.watch_list:
            m.pull(server, collection)
        lf = LeafFile(os.path.join(state_dir, "leaves"))
        new_leaves = m.leaves()
        send_changes_between(lf.get_leaves(), new_leaves, m, c)
        lf.set_leaves(new_leaves)
    finally:
        os.unlink(lockfile)

if __name__ == "__main__":
    main(sys.argv[0], sys.argv[1:])



#!/usr/bin/env python2
#
# Copyright (C) Nathaniel Smith <njs@pobox.com>
#               Timothy Brownawell <tbrownaw@gmail.com>
#               Thomas Moschny <thomas.moschny@gmx.de>
#               Richard Levitte <richard@levitte.org>
# Licensed under the MIT license:
#   http://www.opensource.org/licenses/mit-license.html
# I.e., do what you like, but keep copyright and there's NO WARRANTY.
#
# CIA bot client script for Monotone repositories, written in python.  This
# generates commit messages using CIA's XML commit format, and can deliver
# them using either XML-RPC or email.  Based on the script 'ciabot_svn.py' by
# Micah Dowty <micah@navi.cx>.

# This version is modified to be called by a server hook, instead of a cron job.

# To use:
#   -- create a configuration file called ciabot.conf in the server's
#      configuration directory (whatever is returned by get_confdir()).
#      That configuration file follows basic-io syntax, the comments in
#      the config class is the best documentation for now.
#   -- include monotone-ciabot.lua in the server's monotonerc
#   -- add a variable 'ciabot_python_script in the server's monotonerc,
#      that points to this script.

import sys
import re
import os
import fnmatch

TOKEN = re.compile(r'''
    "(?P<str>(\\\\|\\"|[^"])*)"
    |\[(?P<id>[a-f0-9]{40}|)\]
    |(?P<key>\w+)
    |(?P<ws>\s+)
''', re.VERBOSE)

def parse_basic_io(raw):
    parsed = []
    key = None
    for m in TOKEN.finditer(raw):
        if m.lastgroup == 'key':
            if key:
                parsed.append((key, values))
            key = m.group('key')
            values = []
        elif m.lastgroup == 'id':
            values.append(m.group('id'))
        elif m.lastgroup == 'str':
            value = m.group('str')
            # dequote: replace \" with "
            value = re.sub(r'\\"', '"', value)
            # dequote: replace \\ with \
            value = re.sub(r'\\\\', r'\\', value)
            values.append(value)
    if key:
        parsed.append((key, values))
    return parsed

class config:
    # The server to deliver XML-RPC messages to, if using XML-RPC delivery.
    # In the configuration file, this is set with the key 'xmlrpc_server'
    xmlrpc_server = "http://cia.navi.cx"

    # The 'from' address to put on email, if using email delivery.
    # In the configuration file, this is set with the key 'from_address'
    from_address = "cia-user@FIXME"

    # The email address to deliver messages to, if using email delivery.
    # In the configuration file, this is set with the key 'to_address'
    smtp_address = "cia@cia.navi.cx"

    # The SMTP server to connect to, if using email delivery.
    # In the configuration file, this is set with the key 'smtp_server'
    smtp_server = "localhost"

    # Set to one of "xmlrpc", "email", "debug".
    # In the configuration file, this is set with the key 'delivery'
    delivery = "debug"

    # The default URL template to be used when none other is found.
    # In the configuration file, this is set with the key 'url'.
    # NOTE: THIS MUST COME BEFORE ANY 'pattern'
    # The URL template can contain %s constructs as follows:
    #   %(revision)s      is replaced with the revision identity
    #   %(branch)s        is replaced with the branch name
    #   %(projectid)s     is replaced with the project identity
    #   %(projectname)s   is replaced with the project name
    default_url = None
    
    # These are three dicionaries, where the first stores project identities
    # keyed with globs and the others store project names for CIA and commit
    # URLs, both keyed with project identities.
    # In the configuration file, this is set with stanzas started with
    # the key(s) 'pattern' followed by the keys 'projectid' (project id),
    # 'projectname' (project name) and 'url' (URL template) to set the
    # values.  Each stanza can have several patterns but only one of
    # 'projectid', 'projectname' and 'url'.
    # You MUST define a project identity in each stanza.
    # Project names are names that are used officially with CIA.  if
    # undefined, the project identity is used.
    projectids = {}
    projectnames = {}
    urls = {}

    # This is internal, an array to keep the patterns in order.
    # THIS MEANS THE ORDER IN THE CONFIGURATION FILE IS SIGNIFICANT!
    patterns = []

    def __init__(self, config_file):
        s = ""
        f = open(config_file)
        for line in f:
            s = s + line
        f.close()
        previous_key = ""
        current_patterns = []
        for key, value in parse_basic_io(s):
            if key == 'xmlrpc_server':
                self.xmlrcp_server = value[0]
            elif key == 'to_address':
                self.smtp_address = value[0]
            elif key == 'from_address':
                self.from_address = value[0]
            elif key == 'smtp_server':
                self.smtp_server = value[0]
            elif key == 'delivery':
                self.delivery = value[0]
            elif key == 'pattern':
                if previous_key != key:
                    # new series of patterns, new stanza
                    current_patterns = []
                self.patterns.append(value[0])
                current_patterns.append(value[0])
            elif key == 'projectid':
                for p in current_patterns:
                    self.projectids[p] = value[0]
            elif key == 'projectname':
                for p in current_patterns:
                    self.projectnames[p] = value[0]
            elif key == 'url':
                url_set = False
                for p in current_patterns:
                    self.urls[p] = value[0]
                    url_set = True
                if not url_set:
                    # It means we haven't seen any pattern yet, and
                    # therefore have a default URL
                    self.default_url = value[0]
            previous_key = key

    def _pattern_for_branch(self, branchname):
        l = [ p for p in self.patterns if fnmatch.fnmatchcase(branchname, p) ]
        if len(l) > 0:
            return l[0]
        return None
        
    def projectid_for_branch(self, branchname):
        pat = self._pattern_for_branch(branchname)
        if pat:
            return self.projectids[pat]
        return None

    def projectname_for_branch(self, branchname):
        pat = self._pattern_for_branch(branchname)
        if pat:
            return self.projectnames[pat] if pat in self.projectnames else self.projectids[pat]
        return None

    def url_for_revision(self, branchname, revid):
        pat = self._pattern_for_branch(branchname)
        if pat:
            substs = {}
            substs["revision"] = revid;
            substs["branch"] = branchname;
            substs["projectid"] = self.projectids[pat];
            substs["projectname"] = self.projectnames[pat] if pat in self.projectnames else self.projectids[pat]
            return (self.urls[pat] if pat in self.urls else self.default_url) % substs
        return ""

################################################################################

def escape_for_xml(text, is_attrib=0):
    text = text.replace("&", "&amp;")
    text = text.replace("<", "&lt;")
    text = text.replace(">", "&gt;")
    if is_attrib:
        text = text.replace("'", "&apos;")
        text = text.replace("\"", "&quot;")
    return text

def send_message(message, c):
    if c.delivery == "debug":
        print message
    elif c.delivery == "xmlrpc":
        import xmlrpclib
        xmlrpclib.ServerProxy(c.xmlrpc_server).hub.deliver(message)
    elif c.delivery == "email":
        import smtplib
        smtp = smtplib.SMTP(c.smtp_server)
        smtp.sendmail(c.from_address, c.smtp_address,
                      "From: %s\r\nTo: %s\r\n"
                      "Subject: DeliverXML\r\n\r\n%s"
                      % (c.from_address, c.smtp_address, message))
    else:
        sys.exit("delivery option must be one of 'debug', 'xmlrpc', 'email'")

def send_change_for(rid, branch, author, log, rev, c):
    message_tmpl = """<message>
    <generator>
        <name>Monotone CIA Bot client python script</name>
        <version>1.0</version>
    </generator>
    <source>
        <project>%(project)s</project>
        <branch>%(branch)s</branch>
    </source>
    <body>
        <commit>
            <revision>%(rid)s</revision>
            <author>%(author)s</author>
            <files>%(files)s</files>
            <log>%(log)s</log>
            %(extra)s
        </commit>
    </body>
</message>"""
    
    substs = {}
    files = []
    for key, values in parse_basic_io(rev):
        if key == 'old_revision':
            # start a new changeset
            oldpath = None
        if key == 'delete':
            files.append('<file action="remove">%s</file>'
                         % escape_for_xml(values[0]))
        elif key == 'rename':
            oldpath = values[0]
        elif key == 'to':
            if oldpath:
                files.append('<file action="rename" to="%s">%s</file>'
                             % (escape_for_xml(values[0]), escape_for_xml(oldpath)))
                oldpath = None
        elif key == 'add_dir':
            files.append('<file action="add">%s</file>'
                         % escape_for_xml(values[0] + '/'))
        elif key == 'add_file':
            files.append('<file action="add">%s</file>'
                          % escape_for_xml(values[0]))
        elif key == 'patch':
            files.append('<file action="modify">%s</file>'
                         % escape_for_xml(values[0]))
            
    substs["files"] = "\n".join(files)
    changelog = log.strip()
    project = c.projectname_for_branch(branch)
    if project is None:
        return
    commiturl = c.url_for_revision(branch, rid)
    substs["author"] = escape_for_xml(author)
    substs["project"] = escape_for_xml(project)
    substs["branch"] = escape_for_xml(branch)
    substs["rid"] = escape_for_xml(rid)
    substs["log"] = escape_for_xml(changelog)
    substs["extra"] = ""
    if not commiturl is None:
        substs["extra"] += '<url>%s</url>' % escape_for_xml(commiturl)

    message = message_tmpl % substs
    send_message(message, c)

def main(progname, args):
    if len(args) != 6:
        sys.exit("Usage: %s confdir revid branch author changelog revision_text" % (progname, ))
    # We don't want to clutter the process table with zombies; but we also
    # don't want to force the monotone server to wait around while we call the
    # CIA server.  So we fork -- the original process exits immediately, and
    # the child continues (orphaned, so it will eventually be reaped by init).
    if hasattr(os, "fork"):
        if os.fork():
            return
    (confdir, rid, branch, author, log, rev, ) = args
    c = config(confdir + "/ciabot.conf")
    send_change_for(rid, branch, author, log, rev, c)

if __name__ == "__main__":
    main(sys.argv[0], sys.argv[1:])



#!/usr/bin/python

# to get calltree on stdout, feed to this file's stdin, the output of:
#
#----------------
#
# #!/usr/sbin/dtrace -qs
# syscall:::entry
# /pid == $target/
# {
#  self->start = timestamp;
# }
# syscall:::return
# /self->start/
# {
#  now=timestamp;
#  printf("syscall: %d\t%s\n", now - self->start, probefunc);
#  ustack();
#  self->start = 0;
# }
# 
#----------------
#
# It looks like:
#
#----------------
# syscall: 47029 munmap
#               ld.so.1`munmap+0x7
#               ld.so.1`leave+0x83
#               ld.so.1`call_init+0x41
#               ld.so.1`setup+0xe2c
#               ld.so.1`_setup+0x28c
#               ld.so.1`_rt_boot+0x56
#               0x80473dc
#
# syscall: 20017 mmap
#               libc.so.1`mmap+0x15
#               libc.so.1`lmalloc+0x6c
#               libc.so.1`atexit+0x1c
#               libc.so.1`libc_init+0x40
#               ld.so.1`call_init+0x46
#               ld.so.1`setup+0xe2c
#               ld.so.1`_setup+0x28c
#               ld.so.1`_rt_boot+0x56
#               0x80473dc
#
# syscall: 4092  fstat64
#               libc.so.1`fstat64+0x15
#               libc.so.1`opendir+0x3e
#               ls`0x8052605
#               ls`0x8051b2b
#               ls`main+0x637
#               ls`0x8051106
#
# syscall: 1234  fstat64
#               libc.so.1`mmap+0x15
#               libc.so.1`lmalloc+0x6c
#               libc.so.1`atexit+0x1c
#               libc.so.1`libc_init+0x40
#               ld.so.1`call_init+0x46
#               ld.so.1`setup+0xe2c
#               ld.so.1`_setup+0x28c
#               ld.so.1`_rt_boot+0x56
#               0x80473dc
#
#----------------

# See also:
#   http://kcachegrind.sourceforge.net/cgi-bin/show.cgi/KcacheGrindCalltreeFormat

import sys
import os.path

def main():
    data = read(sys.stdin)
    data.writeto(sys.stdout)

# We need to eventually get, for each function:
#   -- time spent directly in it
#   -- list of functions it calls
#      -- call count (which we don't have)
#      -- the file/line number/name of the function called
#      -- inclusive cost of those calls
# To do this, we have to accumulate everything in memory, then write it out.

class FnEntry:
    def __init__(self, obj_symbol):
        (self.obj, self.symbol) = obj_symbol
        self.cost = 0
        self.children = {}

    def charge(self, cost):
        self.cost += cost

    def charge_child(self, obj_symbol, cost):
        self.children.setdefault(obj_symbol, 0)
        self.children[obj_symbol] += cost

    def writeto(self, stream):
        stream.write("ob=%s\n" % self.obj)
        stream.write("fn=%s\n" % self.symbol)
        # <line> <direct cost>
        stream.write("%s %s\n" % (1, self.cost))
        for ((obj, symbol), cost) in self.children.iteritems():
            stream.write("cob=%s\n" % obj)
            stream.write("cfn=%s\n" % symbol)
            # <number calls> <location in file>
            stream.write("calls=1 1\n")
            # <caller line> <cost>
            stream.write("%s %s\n" % (1, cost))
        stream.write("\n")

class Data:
    def __init__(self):
        self.fns = {}

    def get_entry(self, obj_symbol):
        if obj_symbol not in self.fns:
            self.fns[obj_symbol] = FnEntry(obj_symbol)
        return self.fns[obj_symbol]
    
    # stack is like [("libc.so.1", "fstat64"), ("ls", "main")], with outermost
    # function last
    def charge(self, cost, stack):
        assert stack
        prev = None
        for obj_symbol in stack:
            entry = self.get_entry(obj_symbol)
            if prev is None:
                entry.charge(cost)
            else:
                entry.charge_child(prev, cost)
            prev = obj_symbol
                
    def writeto(self, stream):
        stream.write("events: syscall_wallclock\n")
        stream.write("\n")
        for entry in self.fns.itervalues():
            entry.writeto(stream)

def read_stack(stream):
    stack = []
    for line in stream:
        line = line.strip()
        if not line:
            break
        if "`" in line:
            obj, rest = line.split("`", 1)
            # libc.so.1`fstat64+0x15
            if "+" in rest:
                symbol, rest = rest.split("+", 1)
            # ls`0x8052605
            else:
                symbol = rest
            stack.append((obj, symbol))
        else:
            # 0x80473dc
            stack.append(("?", line))
    return stack

def read(stream):
    data = Data()
    for line in stream:
        if not line.startswith("syscall:"):
            # skip over nonsense
            continue
        else:
            # we have the beginning of a sample
            syscall_marker, cost_str, syscall = line.strip().split()
            assert syscall_marker == "syscall:"
            cost = int(cost_str)
            stack = read_stack(stream)
            data.charge(cost, [("kernel", syscall)] + stack)
    return data

if __name__ == "__main__":
    main()



#!/usr/bin/python

# feed this file opreport -gcf on stdin, and it will spit out
# calltree on stdout

# http://kcachegrind.sourceforge.net/cgi-bin/show.cgi/KcacheGrindCalltreeFormat

import sys
import os.path

def main():
    convert(sys.stdin, sys.stdout)

# need to store:
#   bunch of function data
#   for each function, its file/line location, plus name
#     plus total samples directly in it
#     plus list of functions called
#       -- call count (which we don't have)
#       -- the file/line number/name of the function called
#       -- inclusive cost of those calls
def convert(op, ct):
    for line in op:
        if line.startswith("Counted "):
            ct.write("events: %s" % line.split(" ", 3)[1])
            continue
        if line.startswith("-" * 70):
            # found our first stanza
            break
    stanza = []
    for line in op:
        if line.startswith("-" * 70):
            process_stanza(stanza, ct)
            stanza = []
        else:
            stanza.append(line)

# some sample lines:
# 601802    9.9042  botan/sha160.cpp:54         /home/njs/src/monotone/vlogs-test/mtn-client _ZN5Botan7SHA_1604hashEPKh
# 1585865  26.0995  (no location information)   /usr/lib/libstdc++.so.6.0.7 (no symbols)
# we return the tuple: (count, filename, line, object, symbol)
def parse_line(line):
    rest = line.strip()
    count, percent, rest = rest.split(None, 2)
    if rest.startswith("(no location information)"):
        filename = "(unknown)"
        line = "0"
        blah, blah, blah, rest = rest.split(None, 3)
    else:
        filename_line, rest = rest.split(None, 1)
        assert ":" in filename_line
        filename, line = filename_line.split(":")
    object, rest = rest.split(None, 1)
    symbol = rest.strip()
    return (count, filename, line, object, symbol)

def process_stanza(stanza, ct):
    ct.write("\n")
    it = iter(stanza)
    # skip over the parent call info
    for line in it:
        # skip past the caller lines
        if line[0] not in "0123456789":
            continue
        # process the direct cost line
        count, filename, line, object, symbol = parse_line(line)
        ct.write("fl=%s\n" % filename)
        ct.write("ob=%s\n" % object)
        ct.write("fn=%s\n" % symbol)
        ct.write("%s %s\n" % (line, count))
        # save for later
        caller_line = line
        # start processing next set of lines
        break
    # process the cumulative child cost lines
    for line in it:
        if "[self]" in line:
            continue
        count, filename, line, object, symbol = parse_line(line)
        ct.write("cfi=%s\n" % filename)
        ct.write("cob=%s\n" % object)
        ct.write("cfn=%s\n" % symbol)
        # we don't know how many calls were made, so just hard-code to "1"
        # and we don't know the line number the calls were made from, so just
        # pretend everything came from the same line our function started on
        ct.write("calls=%s %s\n" % (1, line))
        ct.write("%s %s\n" % (caller_line, count))

if __name__ == "__main__":
    main()
